; SOUND.ASM
; ---------
;

	
PSG_ENABLE	= $80
PSG_DDA		= $40
PSG_NOISE	= $80
PSG_LFO_ON	= $80


SND_STOP	= 0	; stop processing (track finish)
			; 0-byte operand

SND_OFF		= 1	; mute sound
			; 0-byte operand

SND_ON		= 2	; turn on sound
			; 0-byte operand

SND_VOLUME	= 3	; set voice volume (left & right)
			; 1-byte operand

SND_FREQ	= 4	; set voice frequency
			; 2-byte operand

SND_DURATION	= 5	; set voice duration
			; 1-byte operand

SND_NOISE_FREQ	= 6	; set noise frequency
			; 1-byte operand

SND_NOISE_OFF	= 7	; set noise off
			; 0-byte operand

SND_LFO_FREQ	= 8	; set LFO frequency
			; 1-byte operand

SND_LFO_CTRL	= 9	; set LFO control
			; 1-byte operand

SND_WAVE_SINE	= 20	; init to sine wave (predefined)
			; 0-byte operand

SND_WAVE_SAW	= 21	; init to sawtooth wave (predefined)
			; 0-byte operand

SND_WAVE_SQR	= 22	; init to square wave (predefined)
			; 0-byte operand

SND_WAVE_DATA	= 30	; init to inline wave data
			; 32-byte operand

                .list
		.zp

wavptr:		.ds	2
sndtrkptr:	.ds	2


		.bss

sndcount0:	.ds	1
sndcount1:	.ds	1
sndcount2:	.ds	1
sndcount3:	.ds	1
sndcount4:	.ds	1
sndcount5:	.ds	1

sndptr0:	.ds	2
sndptr1:	.ds	2
sndptr2:	.ds	2
sndptr3:	.ds	2
sndptr4:	.ds	2
sndptr5:	.ds	2

snddatabank0:	.ds	1
snddatabank1:	.ds	1
snddatabank2:	.ds	1
snddatabank3:	.ds	1
snddatabank4:	.ds	1
snddatabank5:	.ds	1

sndbanktmp:	.ds	1
		
		.code


                .proc _snd_volume
                stx     psg_mainvol
                rts
                .endp

                .proc _snd_init
;
; initialize sound
;
; snd_init:
                stz     sndptr0
		stz	sndptr1
		stz	sndptr2
		stz	sndptr3
		stz	sndptr4
		stz	sndptr5

		stz	sndcount0
		stz	sndcount1
		stz	sndcount2
		stz	sndcount3
		stz	sndcount4
		stz	sndcount5

		stz	psg_mainvol
		stz	psg_lfoctrl

		lda	#5	; set volume to zero for each channel
.l1:		sta	psg_ch	; and disable them
		stz	psg_ctrl
		stz	psg_pan
		dec	A
		bpl	.l1

		lda	#4	; disable noise for channel 5/6
		sta	psg_ch
		stz	psg_noise
		lda	#5
		sta	psg_ch
		stz	psg_noise
		rts

                .endp

;
; snd_trkreg
; ----
; register a track into a voice
; ----
; IN:  _AL = voice number
;      _BL = data bank 
;      _SI = address of track data
; ----
                .proc _snd_trkreg.2

; snd_trkreg:
                lda     <_al
		tax
		stz	sndcount0,x
		asl	A
		tax
		stw	<_si, sndptr0,x
		lda	<_bl
		sta	snddatabank0,x
		rts
                .endp

;
; snd_interval
; ----
; continue playing the track at a vsync/timer interval,
; updating counters as necessary
; ----
                .proc _snd_interval
; snd_interval:
                ldx     #0
.loop:		lda	sndcount0,x	; 6 downcounters
		beq	.downcnt
		dec	A		; not zero, count down
		sta	sndcount0,x
		jmp	.next

.downcnt:	stx	<_al		; if zero, check if voice has
		asl	<_al		; registered playable track
		stz	<_ah
		stw	#sndptr0,<_cx	; _cx is pointer to track pointer
		addw	<_ax,<_cx

		cly
		lda	[_cx],y		; now sndtrkptr is track pointer
		sta	<sndtrkptr
		iny
		lda	[_cx],y
		sta	<sndtrkptr+1

		cmpw	#0,<sndtrkptr		; if track pointer is zero
		beq	.next		; (no track), go to next voice

		stw	<sndtrkptr,<_si
		lda	snddatabank0,x
		sta	<_bl
                phx
		jsr	map_data
                plx
		stw	<_si,<sndtrkptr
		
.exe:		phx			; else execute sound
		stx	<_al
		txa
		sta	psg_ch		; set channel

;                pha
;                tam     #4
;                sta     sndbanktmp
;                lda     BANK(snd_execute)
;                tma     #4
;                pla
	
                call     snd_execute     ; interpret byte-stream

;                pha
;                lda     sndbanktmp
;                tam     #4
;                pla
	
		plx

		stz	<_ah		; at end, update track pointer
		asl	<_al		; in voice context
		addw	#sndptr0,<_ax
		lda	<sndtrkptr
		cly
		sta	[_ax],y
		iny
		lda	<sndtrkptr+1
		sta	[_ax],y

		jsr	unmap_data
	
.next:		inx
		cpx	#6
		beq	.exit
		jmp	.loop
.exit:	
		rts
                .endp        


;                .bank   SOUND_BANK
;                .org    $8000
	
;
; snd_execute
; ----
; execute the sound track currently registered
; ----
; IN: _AL = voice
;     sndtrkptr = ptr to sound code
; ----
                .proc snd_execute
; snd_execute:

;.loop:		cly
;		lda	[sndtrkptr]
;
;		cmp	#SND_STOP
;		bne	.byte1
;		cla
;		sta	psg_ctrl
;		sta	<sndtrkptr
;		sta	<sndtrkptr+1
;		jmp	.out
;
;.byte1:		cmp	#SND_OFF
;		bne	.byte2
;		stz	psg_ctrl	; off with no volume
;		incw	<sndtrkptr
;		jmp	.loop
;
;.byte2:		cmp	#SND_ON
;		bne	.byte3
;		lda	#(PSG_ENABLE | $1F)	; on + full volume
;		sta	psg_ctrl
;		incw	<sndtrkptr
;		jmp	.loop
;
;.byte3:		cmp	#SND_VOLUME
;		bne	.byte4
;		iny
;		lda	[sndtrkptr],y
;		sta	psg_pan
;		addw	#2,<sndtrkptr
;		jmp	.loop
;
;.byte4:		cmp	#SND_FREQ
;		bne	.byte5
;		iny
;		lda	[sndtrkptr],y
;		sta	psg_freqlo
;		iny
;		lda	[sndtrkptr],y
;		sta	psg_freqhi
;		addw	#3,<sndtrkptr
;		jmp	.loop
;
;.byte5:		cmp	#SND_DURATION
;		bne	.byte6
;		lda	<_al
;		tax
;		iny
;		lda	[sndtrkptr],y
;		sta	sndcount0,x
;		addw	#2,<sndtrkptr
;		jmp	.out
;
;.byte6:		cmp	#SND_NOISE_FREQ
;		bne	.byte7
;		iny
;		lda	[sndtrkptr],y
;		sta	psg_noise
;		addw	#2,<sndtrkptr
;		jmp	.loop
;
;.byte7:		cmp	#SND_NOISE_OFF
;		bne	.byte8
;		iny
;		stz	psg_noise
;		incw	<sndtrkptr
;		jmp	.loop
;
;.byte8:		cmp	#SND_LFO_FREQ
;		bne	.byte9
;		iny
;		lda	[sndtrkptr],y
;		sta	psg_lfofreq
;		addw	#2,<sndtrkptr
;		jmp	.loop
;
;.byte9:		cmp	#SND_LFO_CTRL
;		bne	.byte20
;		iny
;		lda	[sndtrkptr],y
;		sta	psg_lfoctrl
;		addw	#2,<sndtrkptr
;		jmp	.loop
;
;.byte20:	cmp	#SND_WAVE_SINE
;		bne	.byte21
;		stw	#sine,<wavptr
;		jsr	snd_loadwave
;		incw	<sndtrkptr
;		jmp	.loop
;
;.byte21:	cmp	#SND_WAVE_SAW
;		bne	.byte22
;		stw	#sawtooth,<wavptr
;		jsr	snd_loadwave
;		incw	<sndtrkptr
;		jmp	.loop
;
;.byte22:	cmp	#SND_WAVE_SQR
;		bne	.byte30
;		stw	#square,<wavptr
;		jsr	snd_loadwave
;		incw	<sndtrkptr
;		jmp	.loop
;
;.byte30:	cmp	#SND_WAVE_DATA
;		bne	.out
;		incw	<sndtrkptr
;		stw	<sndtrkptr,<wavptr
;		jsr	snd_loadwave
;		stw	<wavptr,<sndtrkptr
;		jmp	.loop
;
;.out:		rts

;
; snd_execute2
; ----
; execute the sound track currently registered with the new opcode set
; ----
; IN: _AL = voice
;     sndtrkptr = ptr to sound code
; ----

MAX_INDEX_NOTE          =  96
MAX_INDEX_FX            = 112
MAX_INDEX_EXTENDED_FX   = 128
MAX_INDEX_DURATION      = 192
MAX_INDEX_VOLUME        = 224
MAX_INDEX_WAVEFORM      = 256

INDEX_VOLUME            = MAX_INDEX_NOTE + $C


;snd_execute2:
.loop:		cly
                ; TEST
                ; sty     <_al
                ; sty     psg_ch          ; set channel
		; TEST
		lda	[sndtrkptr]

		cmp	#MAX_INDEX_NOTE
                bhs     .not_note
		tay
		lda 	note_lo,y
		sta	psg_freqlo
		lda	note_hi,y
		sta	psg_freqhi
		incw	<sndtrkptr
		jmp 	.loop

.not_note:
		cmp	#MAX_INDEX_FX
                bhs     .not_fx
	
		cmp	#INDEX_VOLUME
		bne	.not_fx_volume

		iny
		lda	[sndtrkptr],y
		ora	#PSG_ENABLE
		sta	psg_ctrl
		addw	#2,<sndtrkptr
		jmp 	.loop

.not_fx_volume:
.not_fx_unknown:
		addw	#2,<sndtrkptr
		jmp 	.loop
		
		
.not_fx:

		cmp	#MAX_INDEX_EXTENDED_FX
                bhs     .not_efx

.not_efx_unknown:
		addw	#2,<sndtrkptr
		jmp	.loop

.not_efx	
		cmp	#MAX_INDEX_DURATION
                bhs     .not_duration

		ldx	<_al

                sub     #MAX_INDEX_EXTENDED_FX
		sta	sndcount0,x

		incw	<sndtrkptr

		jmp	.out

.not_duration
		cmp 	#MAX_INDEX_VOLUME
                bhs     .not_volume

                sub     #MAX_INDEX_DURATION
		ora	#PSG_ENABLE

		sta	psg_ctrl
		incw	<sndtrkptr
		jmp	.loop		

.not_volume
		stw	#waveform0,<wavptr
		
		lda	[sndtrkptr]
                sub     #MAX_INDEX_VOLUME
		clx

		.ifdef HUC

		__pushw
		cla
		ldx	#5
			; we're shifting on the left the index of the waveform by 5
		jsr 	asl
		
		__addw	<wavptr
			; now, wavptr should point to waveformXX where XX is the arg of the
			; waveform opcode

		.else
		  ; define index shifting with asm non-huc only 
		.endif

		jsr	snd_loadwave

		incw	<sndtrkptr
		jmp	.loop

		stw	<sndtrkptr,<wavptr		
		

.out		rts

note_lo:	
	db 187,058,208,122,055,006,230,214,215,228,000,041
	db 093,157,232,061,155,003,115,235,107,242,128,020
	db 175,078,244,158,078,002,185,118,054,249,192,138
	db 087,039,250,207,167,129,093,059,027,253,224,197
	db 172,148,125,104,083,064,046,029,013,254,240,227
	db 214,202,190,180,170,160,151,143,135,127,120,113
	db 107,101,095,090,085,080,076,071,067,064,060,057
	db 053,050,048,045,042,040,038,036,034,032,030,028

note_hi:
	db 026,025,023,022,021,020,018,017,016,015,015,014
	db 013,012,011,011,010,010,009,008,008,007,007,007
	db 006,006,005,005,005,005,004,004,004,003,003,003
	db 003,003,002,002,002,002,002,002,002,001,001,001
	db 001,001,001,001,001,001,001,001,001,000,000,000
	db 000,000,000,000,000,000,000,000,000,000,000,000
	db 000,000,000,000,000,000,000,000,000,000,000,000
	db 000,000,000,000,000,000,000,000,000,000,000,000


;
; load a waveform into a PSG channel
; PSG channel must already be set
;
snd_loadwave:	lda	#0
		sta	psg_ctrl

		ldx	#$20
.l1:		lda	[wavptr]
		sta	psg_wave
		incw	<wavptr
		dex
		bne	.l1

		lda	#PSG_ENABLE
		ora	#$1F		; full volume
		sta	psg_ctrl
		lda	#$ff
		sta	psg_pan
		rts
		
;
; standard wave tables
;
waveform0:
sine:		db	18,22,24,26,28,28,30,30
		db	30,30,28,28,26,24,22,18
		db	12, 8, 6, 4, 2, 2, 0, 0
		db	 0, 0, 2, 2, 4, 6, 8,12
sawtooth:	db	17,19,21,23,25,27,29,31
		db	31,29,27,25,23,21,19,17
		db	15,13,11, 9, 7, 5, 3, 1
		db	 1, 3, 5, 7, 9,11,13,15
square:		db	 0, 0, 0, 0, 0, 0, 0, 0
		db	 0, 0, 0, 0, 0, 0, 0, 0
		db	31,31,31,31,31,31,31,31
		db	31,31,31,31,31,31,31,31

                .endp

